#!/usr/bin/env python
import os
import struct

import es

################################################################################
## Constants
gamePath = str(es.ServerVar('eventscripts_gamedir'))
gamePath = gamePath.replace('\\', '/')

################################################################################
## SourceModel class
class SourceModel(object):
    '''!SourceModel class is used for opening and manipulating Source-formatted
    MDL files.'''
    
    ## C Struct header of a MDL file (taken directly from the Source SDK sourcecode @ studio.h, studiohdr_t)
    STRUCT_FORMAT_LIST = [
        'i',    # ID - 0
        'i',    # Version - 1
        'l',    # Checksum - 2
        '64s',  # Model name - 3
        'i',    # Length - 4
        
        ('x' * 12), # Vector
        ('x' * 12), # Vector
        ('x' * 12), # Vector
        ('x' * 12), # Vector
        ('x' * 12), # Vector
        ('x' * 12), # Vector
        
        'i',    # Flags - 5
        
        'i',    # Number of bones - 6
        'i',    # Bone index - 7
        
        'i',    # Number of bone controllers - 8
        'i',    # Bone controller index - 9
        
        'i',    # Number of hitbox sets - 10
        'i',    # Hitbox set index - 11
        
        'i',    # Number of local animations - 12
        'i',    # Local animation index - 13
        'i',    # Number of local sequences - 14
        'i',    # Local sequence index - 15
        'i',    # Activity list version - 16
        'i',    # Events indexed - 17
        
        'i',    # Number of textures - 18
        'i',    # Texture index - 19
        
        'i',    # Number of texture search paths - 20
        'i',    # Texture search path index - 21
        
        # There is more but rather unuseful in Python
    ]
    
    def __init__(self, path, fixPathMismatch=True, debug=False):
        '''!Initializes all paths, opens and parses the file.
        
        @param path Path of the model in question relative to the models directory in the mod root.
        @param fixPathMismatch Whether to automatically fix path mismatches between the MDL and actual.
        @param debug Whether to spew debug messages to the console.'''
        ## Whether to spew debug messages to the console
        self.debug = debug
        
        # Strip the models/ prefix
        if path.startswith('models/'):
            path = path[7:]
        
        # Ensure there is a .mdl ending
        if not path.endswith('.mdl'):
            raise ValueError('SourceModel cannot examine a non-MDL file.')
        
        ## Full file path. enginePath is the path reported by the MDL file, may not also be the same (invalid model if not, use fixModelPath to correct)
        self.filePath = getGameDir('models/%s' % path)
        
        #
        # Reset header data
        #
        ## Unpacked header data. In the same format as the header struct
        self.headerData = None
        self.enginePath = '?? %s ??' % self.filePath
        
        #
        # Set the file struct format and size
        #
        ## Struct header in a string, ready to be passed to struct.[un]pack
        self.STRUCT_FORMAT = ''.join(SourceModel.STRUCT_FORMAT_LIST)
        
        ## Size of the struct
        self.STRUCT_FORMAT_SIZE = struct.calcsize(self.STRUCT_FORMAT)
        
        # Parse the file
        self.__parseFile()
        
        # Do the actual file path and the engine path match?
        if self.filePath != self.fullPath:
            es.dbgmsg(0, ' *** Warning *** Mismatch file path reported by the MDL file and actual file path')
            es.dbgmsg(0, '                 > Actual file path:  %s' % self.filePath)
            es.dbgmsg(0, '                 > Reported model path:  %s' % self.fullPath)
            
            # Fix the file path automatically
            if fixPathMismatch:
                fixModelPath(self.filePath, True)
                
                # Reparse the file
                self.__parseFile()
    
    def __repr__(self):
        '''!Literal string representation of this object, including the model
        path for debugging purposes.'''
        return object.__repr__(self)[:-1] + ' modelpath="%s">' % self.getEnginePath()
    
    def __parsePaths(self):
        '''!Parses the path variables. For use when the file has to be reparsed after model paths have changed.'''
        #
        # Path variables
        #
        modelPath = self.headerData[3]
        
        ## Model path, relative to mod root
        self.enginePath = 'models/%s' % (modelPath[:modelPath.find('\x00')].replace('\\', '/'))
        
        ## Directory of which the model is in, relative to mod root
        self.engineDirectory = self.getEnginePath().rsplit('/', 1)[0]
        
        ## Model path, absolute
        self.fullPath = getGameDir(self.getEnginePath())
        
        ## Directory of which the model is in, absolute
        self.fullDirectory = self.getFullPath().rsplit('/', 1)[0]
    
    def __parseFile(self):
        '''!Parses this MDL file. Should only be called once unless the model
        file has changed.'''
        #
        # Open file and read it
        #
        ## Python file handle to the model
        self.file = open(self.filePath, 'rb')
        
        ## File data (binary)
        self.fileData = self.file.read()
        
        ## File data size (in bytes)
        self.fileDataSize = len(self.fileData)
        
        #
        # Unpack data
        #
        self.headerData = struct.unpack(self.STRUCT_FORMAT, self.fileData[:self.STRUCT_FORMAT_SIZE])
        
        #
        # Find the amount of \x00 there are from the end of the texture search path data
        #
        textureSearchPathData = self.fileData[self.headerData[21]:]
        nullBytes = 0
        
        for byte in textureSearchPathData[::-1]:    # ::-1 is Python shorthand for "reverse"
            # If this byte is null add it to the NULL bytes count
            if byte == '\x00':
                nullBytes += 1
            
            # Break if we haven't found a NULL byte
            else:
                break
        
        # Debug the amount of NULL bytes we found at EOF
        if self.debug:
            es.dbgmsg(0, ' *** Debug %r *** Found "%d" null bytes at EOF.' % (self, nullBytes))
        
        #
        # Split the texture search path data with null bytes (make sure we remove the null bytes at EOF)
        #
        textureSearchPathParts = textureSearchPathData[:-nullBytes].rsplit('\x00', self.headerData[20])
        temporarySearchPaths = textureSearchPathParts[1:]
        
        ## Texture search paths (see getTextureSearchPaths for more)
        self.textureSearchPaths = []
        
        # Clean path
        for index, path in enumerate(temporarySearchPaths):
            # Make Linux path seperators
            temporarySearchPaths[index] = path.replace('\\', '/')
            
            # Remove trailing slash for neatness
            if temporarySearchPaths[index][~0] == '/':
                temporarySearchPaths[index] = temporarySearchPaths[index][:-1]
            
            # Make sure the path exists
            if not os.path.isdir(getGameDir('materials/%s' % temporarySearchPaths[index])):
                es.dbgmsg(0, ' *** Warning *** Invalid texture search path: "materials/%s"' % temporarySearchPaths[index])
                continue
            
            # Add to search path
            self.textureSearchPaths.append(temporarySearchPaths[index])
        
        # Reparse path variables
        self.__parsePaths()
    
    def getTextureSearchPaths(self):
        '''!@return All the texture search paths. These are the directories which
        are used to find model textures. Relative to materials directory in mod
        root.'''
        return self.textureSearchPaths
    
    def getEnginePath(self):
        '''!@returns The path to the file. Relative to mod root.'''
        return self.enginePath
    
    def getEngineDirectory(self):
        '''!@return The directory of which the model is located. Relative to mod
        root.'''
        return self.engineDirectory
    
    def getFullPath(self):
        '''!@return The absolute path to the actual file.'''
        return self.fullPath
    
    def getFullDirectory(self):
        '''!@return The absolute directory of this model.'''
        return self.fullDirectory
    
    def precache(self):
        '''!Precaches this model with the engine.
        
        @return Index of the model'''
        return es.precachemodel(self.getEnginePath())

################################################################################
## Model helpers
def fixModelPath(modelLocation, isAbsolute=False):
    '''!Fixes the model path in the header of the .MDL file. Used for when you
    move a model to a different location.
    
    @param modelLocation Relative to materials directory in mod root (unless specified in isAbsolute)
    @param isAbsolute Is modelLocation an absolute path?
    '''
    STRUCT_SIMPLE_HEADER_LIST = [
        'i',    # ID - 0
        'i',    # Version - 1
        'l',    # Checksum - 2
        '64s',  # Model name - 3
        'i',    # Length - 4
    ]
    
    # Calculate the header size
    simpleHeaderSize = struct.calcsize(''.join(STRUCT_SIMPLE_HEADER_LIST))
    
    # Get model location
    if not isAbsolute:
        modelLocation = getGameDir('models/%s' % modelLocation)
    
    # Get the relative path
    relativePath = modelLocation.rsplit('models/', 1)[~0]
    
    # Get the current model data and header
    oldModelData = open(modelLocation, 'rb').read()
    oldModelHeader = struct.unpack(''.join(STRUCT_SIMPLE_HEADER_LIST), oldModelData[:simpleHeaderSize])
    
    # Create a new model header
    newModelHeader = struct.pack(''.join(STRUCT_SIMPLE_HEADER_LIST), oldModelHeader[0], oldModelHeader[1], oldModelHeader[2], relativePath, oldModelHeader[4])
    newModelData = newModelHeader + oldModelData[simpleHeaderSize:]
    
    # Write the new model to file
    open(modelLocation, 'wb').write(newModelData)

def addModelToDownloadables(modelLocation):
    '''!Adds all the files that this model depends on (i.e. model files such as
    .vvd, .phy, .vtx and material files such as .vmt and .vtf) to the
    downloadables list so clients can see the model correctly.
    
    @remark Also adds the actual model to downloadables, too.
    
    @param modelLocation Relative to materials directory in mod root.
    '''
    srcmdl = SourceModel(modelLocation)
    
    enginePath = srcmdl.getEngineDirectory()
    fullPath = srcmdl.getFullDirectory()
    modelNameNoExt = os.path.splitext(srcmdl.getEnginePath().rsplit('/', 1)[1])[0]
    
    # Add model to downloadables
    es.stringtable('downloadables', srcmdl.getEnginePath())
    
    for searchPath in srcmdl.getTextureSearchPaths():
        # searchPath is relative to ./materials/
        fullSearchPath = getGameDir('materials/%s' % searchPath)
        
        for file in os.listdir(fullSearchPath):
            fileName, fileExt = os.path.splitext(file)
            
            # Check extension
            if fileExt not in ('.vmt', '.vtf'):
                es.dbgmsg(2, ' *** Debug *** Skipping file "%s" (ext: "%s"): is not a VMT or VTF.' % (file, fileExt))
                continue
            
            # Add to downloadables
            es.dbgmsg(1, ' *** Downloadables *** Adding texture: "%s/%s"' % (searchPath, file))
            es.stringtable('downloadables', 'materials/%s/%s' % (searchPath, file))
    
    # Now add all the .mdl, .vtx, .vvd, etc. files from the model path to the
    # downloadables
    for file in os.listdir(fullPath):
        # Get the first file name (before extension[s])
        firstFileName = file.split('.', 1)[0]
        
        # Get file name and extension
        fileName, fileExt = os.path.splitext(file)
        
        # Skip files that don't have the same name as the model
        if firstFileName != modelNameNoExt:
            es.dbgmsg(2, ' *** Debug *** Skipping file "%s" (ext: "%s"): name mismatch.' % (file, fileExt))
            continue
        
        # Check extension
        if fileExt not in ('.vtx', '.vvd', '.mdl', '.phy', '.ani'):
            es.dbgmsg(2, ' *** Debug *** Skipping file "%s" (ext: "%s"): is not a VMT or VTF.' % (file, fileExt))
            continue
        
        # Add to downloadables
        es.dbgmsg(1, ' *** Downloadables *** Adding model: "%s/%s"' % (enginePath, file))
        es.stringtable('downloadables', '%s/%s' % (enginePath, file))

################################################################################
## Generic helpers
def getGameDir(dir):
    '''!Gets an absolute path to a game directory.
    
    @remark Implicitly replaces \\ with / (linux support)
    
    @param dir Directory to append to the game directory.
    
    @return An absolute path to the game directory plus \p dir.'''
    # Linux path seperators
    dir = dir.replace('\\', '/')
    
    # Return
    return '%s/%s' % (gamePath, dir)